/*
* MPIA - GEIRS project
*
* $Header: https://svn.mpia.de/gulli/geirs/src/branches/rjm/FEditHead.cxx 709 2013-10-29 19:43:10Z mathar $
*
*/

#include "FEditHead.h"

#include <string>
#include <cstring>
#include <vector>
#include <fstream>

#include <unistd.h>

#include <boost/regex.hpp>

using namespace std ;
using namespace CCfits ;

/** Constructor given a FITS file and a template ASCII file.
* It attempts to open the FITS file and to scan the template file.
* Applying the template modifications to the FITS file is left to a subsequent
* call to FEditHead::exec .
* @param[in] fitsname The name of the FITS file to be modified.
* @param[in] fitstpl The name of the ASCII file with the header template lines.
* @since 2012-10-18
* @author Richard J. Mathar
*/
FEditHead::FEditHead(char *fitsname, char *fitstpl) : FITS(fitsname,CCfits::Write), cfgFil(fitstpl), buffer(0)
#ifdef FEDITHEAD_KEEP_FLAG
   , keepDtor(true) 
#endif
{
	tplLins = scanTpl(fitstpl) ;
} /* FEditHead::FEditHead */

/** Constructor given a FITS file and a list of template ASCII files.
* It attempts to open the FITS file and to scan all template files.
* Applying the template modifications to the FITS file is left to a subsequent
* call to FEditHead::exec .
* @param[in] fitsname The name of the FITS file to be modified.
* @param[in] fitstpl The name of the ASCII files with the header template lines.
* @since 2013-10-30
* @author Richard J. Mathar
*/
FEditHead::FEditHead(char *fitsname, const vector<string> fitstpl) 
	: FITS(fitsname,CCfits::Write), cfgFil(), buffer(0)
#ifdef FEDITHEAD_KEEP_FLAG
   , keepDtor(true) 
#endif
{
	for(int f =0 ; f < (int) fitstpl.size() ; f++)
	{
		/* attach name to the reported cfgFil */
		cfgFil += string(" ")+fitstpl[f] ;
	}
	tplLins = scanTpl(fitstpl) ;
} /* FEditHead::FEditHead */

/** Constructor given an old-fashioned GEIRS header stream and a template ASCII file.
* It attempts to split the header stream in 80 bytes pieces and to scan the template file.
* Applying the template modifications to the FITS file is left to a subsequent
* call to FEditHead::exec .
* @param[in] gStream The bytes in the GEIRS header.
*  These are characterized by having the shape of non-terminated 80-byte lines.
* @param[in] glines The number of 80-byte lines in gstream.
* @param[in] fitstpl The name of the ASCII file with the header template lines.
* @param[in] fnam The name of the temporary FITS file to be created.
* @since 2013-07-12
* @author Richard J. Mathar
*/
FEditHead::FEditHead(const char *gstream, int glines, int bitpix, int naxis, long *naxes, char *fitstpl, const char *fnam) 
	: FITS(string(fnam),bitpix,naxis,naxes), cfgFil(fitstpl), buffer(0)
#ifdef FEDITHEAD_KEEP_FLAG
     , keepDtor(false) 
#endif
{
	tplLins = scanTpl(fitstpl) ;
	readStream(gstream,glines) ;
} /* FEditHead::FEditHead */

/** Constructor given an old-fashioned GEIRS header stream and template ASCII files.
* It attempts to split the header stream in 80 bytes pieces and to scan all template files.
* Applying the template modifications to the FITS file is left to a subsequent
* call to FEditHead::exec .
* @param[in] gStream The bytes in the GEIRS header.
*  These are characterized by having the shape of non-terminated 80-byte lines.
* @param[in] glines The number of 80-byte lines in gstream.
* @param[in] fitstpl The names of the ASCII files with the header template lines.
* @param[in] fnam The name of the temporary FITS file to be created.
* @since 2013-07-12
* @author Richard J. Mathar
*/
FEditHead::FEditHead(const char *gstream, int glines, int bitpix, int naxis, long *naxes, const vector<string>fitstpl,
	const char *fnam) 
	: FITS(string(fnam),bitpix,naxis,naxes), cfgFil(), buffer(0)
#ifdef FEDITHEAD_KEEP_FLAG
     , keepDtor(false) 
#endif
{
	for(int f =0 ; f < (int) fitstpl.size() ; f++)
	{
		/* attach name to the reported cfgFil */
		cfgFil += string(" ")+fitstpl[f] ;
	}
	tplLins = scanTpl(fitstpl) ;
	readStream(gstream,glines) ;
} /* FEditHead::FEditHead */

/** Constructor given an old-fashioned GEIRS header stream.
* It attempts to split the header stream in 80 bytes pieces.
* Applying the template modifications to the FITS file is left to a subsequent
* call to FEditHead::exec .
* @param[in] gStream The bytes in the GEIRS header.
*  These are characterized by having the shape of non-terminated 80-byte lines.
* @param[in] fnam The name of the temporary FITS file to be created.
* @since 2013-07-12
* @author Richard J. Mathar
*/
FEditHead::FEditHead(const char *gstream, char *fnam, int glines) 
	: FITS(string(fnam),CCfits::Write), buffer(0)
#ifdef FEDITHEAD_KEEP_FLAG
     , keepDtor(false) 
#endif
{
	/* Divide the total stream length (bytes) through 80 to get the
	* number (count) of lines. Note we divide through 80 because
	* there are no LF in the stream.
	*/
	if ( glines < 0 )
		glines = strlen(gstream)/(FLEN_CARD-1) ;
	readStream(gstream,glines,true) ;
} /* FEditHead::FEditHead */

/** Destructor.
* De-allocates the buffer with the sequential concatenation
* of all the 80-bytes header lines.
* Removes the FITS file if the flag keepDtor was set to false.
*/
FEditHead::~FEditHead()
{
        /* remove the resoures allocated within the ctor
        */
        delete[] buffer ;
	buffer=0 ;
#ifdef FEDITHEAD_KEEP_FLAG
	if ( ! keepDtor)

		unlink(name().c_str()) ;
#endif
} /* FEditHead::dtor */


/**  Execute the modifications of the template file in the header.
* Request of adding or modifying keywords are executed in the order
* of request (in the order of the template file) and the new keywords
* are appended in that order. This basically means that header cards are
* not sorted in any particular way to be 'nice.'
* @param[in] verbose If true, comment on each substitution, deletion etc.
* @since 2012-10-18
* @author Richard J. Mathar
*/
void FEditHead::exec(bool verbose)
{
	/** Do nothing if the requested list of keywords is empty.
	*/
	if ( tplLins.size() == 0 )
		return ;

	/* keep informed whether we are in the primary or any extension HDU
	*/
	bool isExten = true ;
	try
	{
		/* the extended syntax may have placed us into one of the
		* extensions, but otherwise, if there is just a primary HDU,
		* we are ending up in the primary HDU and the currentExtension()
		* will throw a noSuchHDU exception.
		*/
		currentExtension() ;
	}
	catch(FITS::NoSuchHDU ex)
	{
		isExten = false;
	}
	HDU & thisex = isExten ? (HDU &) currentExtension() : (HDU &) pHDU() ;
	thisex.readAllKeys() ;

	/* remember whether any keys were actually changed.
	*/
	bool didSth = false;

	map<String, Keyword*> & allkw = thisex.keyWord() ;

	/* Create a  copy of the existing keywords as we will run with an
	* iterator through the keywords, and this will be confused if the
	* map changes while iterating.
	*/
	map<String, Keyword*> origKw(allkw) ;

	/** Consider in the outer loop each applicable template line.
	*/
	vector<string>::const_iterator tplItr = tplLins.begin() ;
	for (  ; tplItr != tplLins.end() ; tplItr++ )
	{
#if 0
		cout << *tplItr << endl ;
#endif
		if ( (*tplItr).substr(0,1) == "-" )
		{
			/* request for deletion. Trim the leading dash and other whit space
			*/
			string delStr = trimws( (*tplItr).substr(1) ) ;

			/* todo: reparse delStr to catch only the keyword, not the
			* value that might also be specified
			*/

#ifdef DEBUG
			cout << "...delete " << delStr << endl ;
#endif
			const boost::regex prim(delStr) ;

			/* for each header keyword line...
			*/
			map<String, Keyword*>::iterator kitr = origKw.begin() ;
			for ( ; kitr != origKw.end() ; kitr++)
			{
				const string & thisk = (*kitr).first ;
				if ( regex_match( thisk, prim) )
				{
					if (verbose)
						cout << "..deleting " << thisk << " which matches " << delStr << endl ;
					thisex.deleteKey(thisk) ;
					didSth = true;
				}
			}
			
		}
		else if ( (*tplItr).substr(0,1) == ":" || (*tplItr).substr(0,1) == "!" )
		{
			/* request for substitution */
			/* split the template line along the colons
			*/
			vector<string> reg = splitColon(*tplItr) ;
			if ( reg.size() == 2 )
			{
#ifdef DEBUG
				cout << "...substitute " << reg[0] << " " << reg[1] << endl ;
#endif
				const boost::regex prim(reg[0]) ;

				/* for each header keyword line...
				*/
				map<String, Keyword*>::iterator kitr = origKw.begin() ;
				for ( ; kitr != origKw.end() ; kitr++)
				{
					const string & thisk = (*kitr).first ;
#if 0
					cout << thisk << endl ;
#endif
					if ( regex_match( thisk, prim) )
					{
#ifdef DEBUG
						cout << thisk << " matches " << reg[0] << endl ;
#endif
						/* compute the new keyword's name
						*/
						string newkw = regex_replace(thisk, prim, reg[1], boost::match_default | boost::format_sed) ;

						if (verbose)
							cout << "... replaced " << thisk << " by " << newkw << endl ;

						/* if the keyword matches the regular expression reg[0],
						* save value and comment, and replace the keyword
						*/
						string thiskCom = (*kitr).second -> comment() ;
						ValueType thiskTyp = (*kitr).second -> keytype() ;

						// long long valLL ;
						// long valL ;
						int valI ;
						double valD ;
						float valF ;
						complex<double> valC ;
						string valS ;
						bool valB ;

						switch(thiskTyp)
						{
						case Tnull:
							thisex.addKey(newkw,"",thiskCom) ;
							break ;
						case Tlogical:
						case VTlogical:
							(*kitr).second ->value(valB) ;
							thisex.addKey(newkw,valB,thiskCom) ;
							break ;
						case Tstring:
							(*kitr).second ->value(valS) ;
							thisex.addKey(newkw,valS,thiskCom) ;
							break ;
						case Tbit:
						case VTbit:
						case Tbyte:
						case VTbyte:
						case Tushort:
						case VTushort:
						case Tshort:
						case VTshort:
						case Tuint:
						case VTuint:
						case Tint:
						case VTint:
						case Tulong:
						case VTulong:
						case Tlong:
						case VTlong:
						case Tlonglong:
						case VTlonglong:
							(*kitr).second ->value(valI) ;
							thisex.addKey(newkw,valI,thiskCom) ;
							break ;
						case Tfloat:
						case VTfloat:
							(*kitr).second ->value(valF) ;
							thisex.addKey(newkw,valF,thiskCom) ;
							break ;
						case Tdouble:
						case VTdouble:
							(*kitr).second ->value(valD) ;
							thisex.addKey(newkw,valD,thiskCom) ;
							break ;
						case Tcomplex:
						case VTcomplex:
						case Tdblcomplex:
						case VTdblcomplex:
							(*kitr).second ->value(valC) ;
							thisex.addKey(newkw,valC,thiskCom) ;
						}
						

						thisex.deleteKey(thisk) ;
						didSth = true;
					}
				}
			}
			/* do nothing here if there was an insufficient number of colons...
			*/
		}
		else if ( (*tplItr).substr(0,8) == "        " )
		{
			/* comment lines (in the FITS sens), starting with at least 8 blanks.
			*/
			thisex.writeComment( (*tplItr).substr(8) );
			if (verbose)
				cout << "... added comment " << (*tplItr).substr(8) << endl ;
		}
		else
		{
			/* request for appendix (or overwriting)
			*/
			char * card = parse_template( *tplItr ) ;
#ifdef DEBUG
			cout << "...append " << card << endl ;
#endif

			bool insrtd = addHduCard(thisex,card,verbose) ;
			didSth |= insrtd ;

			/* release memory allocated in parse_template()
			*/
			delete[] card ;
		}
	}

	if (didSth)
	{
		/* If any keywords were changed or added, write a HISTORY line
		* keeping track of the template file (superflous probably), and remove the CHECKSUM keyword
		* if present.
		*/

#if 0
		/* Keep the date of the original creation intact...
		* thisex.writeDate() ;
		*/
		time_t now = time(NULL) ;
		string nowLoc = ctime(&now) ;
		if ( getenv("LOGNAME") )
			nowLoc += string(getenv("LOGNAME")) ; 
		if ( getenv("DOMAIN") )
			nowLoc += string(" ")+string(getenv("DOMAIN")) ;

		thisex.writeHistory("Modified by "__FILE__" config "+ cfgFil + " " +nowLoc);
#endif

		/* update the CHECKSUM if needed. DATASUM is not needed...
		*/
		try
		{
			thisex.deleteKey("CHECKSUM") ;
		}
		catch(HDU::NoSuchKeyword ex2)
		{
			/* may throw an exception if the keyword is already
			* absent, in which we're not interested
			*/
		}
	}
} /* FEditHead::exec */

/*!********************************************************* 
* Dump the lines into the local buffer and return its location.
* @param[in] maxLines A cutoff for the number of lines to be returned.
*  A negative number is interpreted as that no upper limit is set.
* @return A pointer to a locally allocated buffer.
*  This will be removed once the instance of the class goes out of scope.
* @since 2013-07-12
*/
char * FEditHead::allLines(int maxLines)
{
	char * card = new char[FLEN_CARD] ;

	/* if no constraint on the number of lines was set, take them
	* all */
	PHDU & phdu = pHDU() ;

	/* is this needed? If not, one can delcare this method const and also the previous phdu
	*/
	phdu.readAllKeys() ;

#if 0
	map<string, Keyword*> & allk = phdu.keyWord() ;

	if ( maxLines < 0 )
		maxLines = allk.size() ;

	/* if the constraint on the number of lines was large, take them
	* all */
	maxLines = min(maxLines, (int)(allk.size()) ) ;
#endif

	/* does not occur in normal use
	*/
	if ( buffer) 
		delete[] buffer ;

	/* see also ffhdr2str() in cfitsio/getkey.c
	*/
	int status =0 ;
	int keysex ;
	fits_get_hdrspace(phdu.fitsPointer(),&keysex, NULL, & status) ;

	if ( maxLines < 0 )
		maxLines = keysex ;

	/* if the constraint on the number of lines was large, take them
	* all */
	maxLines= min(maxLines,keysex) ;

	/* 80 bytes per line plus a final \0
	*/
	buffer = new char[(FLEN_CARD-1)*maxLines+1] ;

#if 0

	/* 80 blanks, enough to padd any string (FITS input line) that is shorter
	* than the exact line length.
	*/
	static const string blankpadd("                                                                                ") ;

	/* loop over all lines to be comitted
	*/
	map<string,Keyword*>::iterator ai= allk.begin() ;
	for(int i=0; i <maxLines ; i++, ai++)
	{
		int keytype ;
		status =0;
		string val ;

		fits_parse_template( incard, card,&keytype, &status) ;

		if ( status)
		{
			// fits_report_error(stderr,status) ;
			card[0] = '\0' ;
		}

		/* blank padding, then truncation to 80 bytes */
		snprintf(&buffer[i*(FLEN_CARD-1)],FLEN_CARD,"%s",((*ai).second+blankpadd).c_str() ) ;

		fprintf(stderr,__FILE__" %d '%s'\n",__LINE__,ai->second->name().c_str() ) ;
		//fprintf(stderr,__FILE__" %d '%s'\n",__LINE__,ai->second->value(val).c_str() ) ;
		fprintf(stderr,__FILE__" %d '%s'\n",__LINE__,ai->second->comment().c_str() ) ;
	}
#endif

	/* card indexing is apparently 1-based in cfitsio, so use i+1 in the argument.
	*/
	for(int i=0; i < maxLines ; i++)
	{
		fits_read_record(phdu.fitsPointer(), i+1, card, &status) ;
		// fprintf(stderr,__FILE__" %d %d '%s'\n",__LINE__,i,card) ;
		snprintf(&buffer[i*(FLEN_CARD-1)],FLEN_CARD,"%-80.80s",card ) ;
	}

	/* this should already have been written by the final output
	* of the loop above (if maxLines>0), so it's actually superfluous
	*/
	buffer[(FLEN_CARD-1)*maxLines] = '\0' ;

	delete[] card ;

	return buffer ;
} /* allLines */

/** Strip leading and trailing white space from a string.
* @param[in] instring The initial string from which nonprintable leading and
*   trailing characters or white space (tabs, blanks) ought to be removed.
* @return The string with leading and trailing blanks, tabs etc removed.
*/
string FEditHead::trimws(const string & instring)
{
	/* this is mad actually more strict, using various iswhite()
	string::size_type head = instring.find_first_not_of(" \t\n\r") ;
	*/
	string::size_type head =0 ;
	for( ; head < instring.length() ; head++)
	{
		/* this is liberal and allows punctuation (dots, commas..) in the names
		if ( isalpha(instring[head]) || isdigit(instring[head]) || ispunct(instring[head]) )
		*/
		if ( isalnum(instring[head]) || ispunct(instring[head]) )
			break ;
	}

	/* return empty string if everything is white space or non-printable
	*/
	if ( head == string::npos)
		return string("") ;
	/*
	string::size_type tail = instring.find_last_not_of(" \t\n\r") ;
	*/
	string::size_type tail = instring.length()-1 ;
	for( ; tail >= 0 ; tail--)
	{
		/* this is liberal and allows punctuation (dots, commas..) in the names
		*/
		if ( isalnum(instring[tail]) || ispunct(instring[tail]) )
			break ;
	}
	return instring.substr(head,tail+1) ;
} /* FEditHead::trimws */

/** Reformat an input string into a standard FITS header line.
* This is an intermediate invocation of the fits_parse_template()
* functionality of cfitsio.
* @param[in] instring
* @return A validated copy of the card line.
*   This char array should be delete[]d by the caller after return.
*/
char * FEditHead::parse_template(const string & instring)
{
	int inlen =  instring.length() ;
	char *incard  = new char[inlen+1];
	incard[inlen] = '\0' ;
	strncpy(incard, instring.c_str(),inlen) ;

	char * card = new char[FLEN_CARD] ;
	int keytype ;
	int status =0;
	fits_parse_template( incard, card,&keytype, &status) ;
	if ( status)
	{
		fits_report_error(stderr,status) ;
		card[0] = '\0' ;
	}

	delete[] incard ;
	return card ;
} /* FEditHead::parse_template */

/** Analyse a header card.
* The header card is decomposed into name, value, comment and type.
* @param[in] card The card to be analysed, typically up to 80 characters. 
* @param[out] keyname Keyword name in the card.
* @param[out] keyval Keyword value in the card.
* @param[out] cmt  Comment after the slash in the card.
* @param[out] dtyp Rough type (string, integer, float, boolean or complex)
* @return The cfitsio error code. 0 if there was no error.
*/
int FEditHead::templateAnal(char * card, char keyname[], char keyval[], char cmt[], char dtyp[])
{
	int status =0;
	int keylen ;

	fits_get_keyname(card, keyname, &keylen, &status) ;
#if 0
	if ( status)
	{
	fprintf(stderr,__FILE__" %d %s\n",__LINE__,card) ;
	fprintf(stderr,__FILE__" %d keynam %d\n",__LINE__,status) ;
	}
#endif
	fits_parse_value(card, keyval, cmt, &status) ;
#if 0
	if ( status)
	{
	fprintf(stderr,__FILE__" %d %s\n",__LINE__,card) ;
	fprintf(stderr,__FILE__" %d keyval %d\n",__LINE__,status) ;
	}
#endif
	fits_get_keytype(keyval, dtyp, &status) ;
#if 0
	if ( status)
	{
	fprintf(stderr,__FILE__" %d %s\n",__LINE__,card) ;
	fprintf(stderr,__FILE__" %d keytyp %d\n",__LINE__,status) ;
	}
#endif
	return status ;
} /* templateAnal */

/**  Scan the template file.
* @param[in] fitstpl The name of the template ASCII file.
*  The file contains six types of lines:
*  - Empty lines and lines starting immediately with a hash (#)
*    are ignored (interpreted as lines that comment the template).
*  - Lines starting with a dash (-) contain a keyword (right after
*    the dash with optional white space in between) that is to be removed from the FITS header. If the keyword
*    starts with <code>HIERARCH</code>, it may consist of blank-separated components,
*    or otherwise it needs to follow the standard 8-letter limitation.
*  - Lines starting with a delimiter, either the colon or the exclamation mark, contain two keywords, separated
*    by another delimiter of the same type. The keyword between the first pair of delimiters
*    is replaced by the keyword in the second pair. This
*    syntax is an extension to the standard cfitsio/fitsio template syntax
*    because this allows to delimit multi-strings as seen with the
*    hierarchical keywords, and on which fedithd will choke if used
*    with its implied rule of the (-)-syntax.
*  - Other lines are interpreted as a keyword, value and comment
*    in the standard FITS style. An equal sign separates
*    keyword and value, and the comment following after a slash (/).
*    Hierarchical keywords are allowed. Conversion to upper case letters
*    occurs on the fly. Comments in long input lines are chopped.
*    Note that if the corresponding keyword is already present in the
*    header, it will be replaced by this new one.
* -  Lines starting with at least 8 blanks are inserted as COMMENT
*    cards into the FITS header, effectively replacing these blanks by
*    the COMMENT keyword. Long lines will be wrapped into the next
*    COMMENT card to fit into the 80-bytes limits of the standard.
* @param[in] raw80 If true, the file fitstpl contains lines of 80 bytes without line feeds
* @return The non-comment lines of the file.
*  The order of the input file (top to bottom) is preserved.
* @since 2012-10-18
* @author Richard J. Mathar
*/
vector<string> FEditHead::scanTpl(const char *fitstpl, const bool raw80) const
{
	vector<string> tplLins ;

	/* try to open the file
	*/
	ifstream infil(fitstpl) ;
	if ( ! infil )
	{
		/* cannot open the template file... return empty line list
		cerr << "Cannot open " << fitstpl << endl ;
		*/
	}
	else if ( raw80)
	{
		char linChar[81] ;
		/* loop over all lines in the input file
		*/
		while( ! infil.eof() )
		{
			infil.get(linChar, 80) ;
			string lin(linChar) ;
			/* ignore comment lines and empty lines */
			if ( lin.find("#") == string::npos && lin.size() )
			{
				/* attach lines that are not empty
				*/
				tplLins.push_back(lin) ;
			}
		}
	}
	else
	{
		/* loop over all lines in the input file
		*/
		while( ! infil.eof() )
		{
			string lin ;
			getline(infil,lin) ;
			/* ignore comment lines and empty lines */
			if ( lin.find("#") == string::npos && lin.size() )
			{
				/* attach lines that are not empty
				*/
				tplLins.push_back(lin) ;
			}
		}
	}
	return tplLins ;
} /* FEditHead::scanTpl */

/**  Scan the template file.
* @param[in] fitstpl The name of the template ASCII file.
*  The contents is the same as documented in the signature with the character pointer.
* @param[in] raw80 If true, the file fitstpl contains lines of 80 bytes without line feeds
* @return The non-comment lines of the file.
*  The order of the input file (top to bottom) is preserved.
* @since 2013-10-17
* @author Richard J. Mathar
*/
vector<string> FEditHead::scanTpl(const string fitstpl, const bool raw80) const
{
	return scanTpl(fitstpl.c_str(), raw80) ;
}

/**  Scan a list of template files.
* The function implies a loop of scans over the template files
* in the order of the vector. This means files later (with higher index)
* in the vector may override the specs of lines in files earlier in the vector.
* @param[in] raw80 If true, the file fitstpl contains lines of 80 bytes without line feeds
* @return The non-comment lines of the file.
*  The order of the input file (top to bottom) is preserved.
* @since 2013-10-17
* @author Richard J. Mathar
*/
vector<string> FEditHead::scanTpl(const vector<string>fitstpl, const bool raw80) const
{
	vector<string> allTplLins ;
	/* loop over the files (components of the vector)
	*/
	for(int f =0 ; f < (int) fitstpl.size() ; f++)
	{
		const vector<string> tplLins = scanTpl(fitstpl[f], raw80) ;
		for( int l =0 ; l < (int) tplLins.size() ; l++)
			allTplLins.push_back( tplLins[l]) ;
	}
	return allTplLins ;
} /* FEditHead::scanTpl */

/** Split a string using colons or exclamaion marks as delimiters.
* @param[in] instring This input has the sequential format :....:.....:.... or !...!...!...
*  Either the colon or the exclamation mark are the local delimiters.
*  This means the first, second and third mark delimit the two strings to
*  be defined. The third mark is optional. If missing, the entire residual
*  part of the input defines the 2nd string.
* @return A vector of 2 strings. The first contains the characters
*  between the first and the second mark, the second between the 2nd
*  and the third mark.
* @since 2012-10-18
* @since 2013-01-29 With exclamation marks working as delimiters
* @author Richard J. Mathar
*/
vector<string> FEditHead::splitColon(const string &instring)
{
	/* Attempt to intepret this as a colon-delimited instring
	*/
	vector<string> spli = splitColon(instring,":") ;

	/* If that search for the colon-delimited parts was sucessful..
	*/
	if ( spli.size() > 0 )
		return spli  ;

	/* Attempt to intpret this as a !-delimited string
	*/
	spli = splitColon(instring,"!") ;
	if ( spli.size() > 0 )
		return spli  ;
	else
		cerr << "Format error: Missing delimiters in " << instring << " .. skipped " << endl ;
	return spli ;
} /* FEditHead::splitColon */

/** Split a string using a specific delimiting string.
* @param[in] instring This input has the sequential format <delimit>....<delimit>.....<delimit>...
*  This means the first, second and third mark delimit the two strings to
*  be defined. The third mark is optional. If missing, the entire residual
*  part of the input defines the 2nd string.
* @param[in] delim The delimiting string.
* @return A vector of 2 strings. The first contains the characters
*  between the first and the second mark, the second between the 2nd
*  and the third mark. If the delimiters were not found, an empty vector
*  is returned.
* @since 2013-01-29
* @author Richard J. Mathar
*/
vector<string> FEditHead::splitColon(const string &instring, const string &delim)
{
	/* find the first, up to three, colons
	*/
	string::size_type head = instring.find_first_of(delim) ;
	/* mid is the position relative to the substring, which is head+1+mid 
	* relative to instring
	*/
	string::size_type mid = instring.substr(head+1).find_first_of(delim) ;

	vector<string> spli ;
	/* continue only if at least two colons are found (assuming the
	* third is implicit by the string length
	*/
	if ( head != string::npos && mid != string::npos)
	{
		spli.push_back( instring.substr(head+1,mid) ) ;
		/* tail is the position relative to the substring, which is head+1+mid 
		* relative to instring
		*/
		string::size_type tail = instring.substr(head+2+mid).find_first_of(delim) ;
		if ( tail != string::npos)
			spli.push_back( instring.substr(head+2+mid,tail) ) ;
		else
			spli.push_back( instring.substr(head+2+mid) ) ;
	}
	/* otherwise return the empty string..
	*/
	return spli ;
} /* FEditHead::splitColon */

/*!**********************************************************
* Put the 80-bytes chunks of the stream into the primary header.
* @param gstream The characters to be fed into the primary header.
*  This stream contains neither line feeds nor 0-bytes.
* @param[in] glines The number of 80-byte blocks in gstream
* @param[in] restr Restrict the lines such that NAXIS keywords are blocked
* @since 2013-07-15
*/
void FEditHead::readStream(const char *gstream, int glines, const bool restr)
{
	/* 81 bytes to manage one header card
	*/
	char card[FLEN_CARD] ;
	card[FLEN_CARD-1] = '\0' ;
	PHDU & phdu = pHDU() ;
	/* loop over all virtual lines in the input stream
	* Q: would we rather compute gline from the \0 in the gstream, automated?
	*/
	// fprintf(stderr,__FILE__" %d %d\n",__LINE__,glines) ;
	for( int l =0 ; l < glines ; l++)
	{
		/* grap characters 80*l up to 80*(l+1)
		*/
		strncpy(card, gstream+(FLEN_CARD-1)*l, FLEN_CARD-1) ;
		if ( ! restr || (
			strncmp(card,"NAXIS",5)
			&& strncmp(card,"SIMPLE",6)
			&& strncmp(card,"BITPIX",6)
			) 
		)
		{
			// fprintf(stderr,__FILE__" %d %d %s\n",__LINE__,l,card) ;
			addHduCard( phdu, card,false ) ;
		}

		if ( strncmp(card,"END",3) == 0 )
			break ;
	}
} /* readStream */

/*!*************************************************
* Insert (or replace) a HDU header card.
* @param hdu The header-data unit to be checked for the card.
* @param card The (first 80 bytes of) the new header card.
* @param verbose
* @return true if the new card was installed.
*    Return values of false occur if the syntax of the card is not valid.
* @since 2013-07-15
*/
bool FEditHead::addHduCard(HDU & hdu, char * card, const bool verbose)
{
	char keyname[FLEN_KEYWORD] ;
	char keyval[FLEN_VALUE] ;
	char cmt[FLEN_COMMENT] ;
	char dtyp[4] ;

	int fclass = fits_get_keyclass(card) ;
	switch ( fclass)
	{
	case TYP_COMM_KEY :
		if ( strncmp("COMMENT ",card,8) == 0 )
			hdu.writeComment( card+8 );
		else if ( strncmp("HISTORY ",card,8) == 0 )
			hdu.writeHistory( card+8);
		return true ;
		break;
	case TYP_STRUC_KEY :
		break;
	}

#if 0
	fits_get_keyname(card, keyname, &keylen, &status) ;
	fits_parse_value(card, keyval, cmt, &status) ;
	fits_get_keytype(keyval, dtyp, &status) ;
#endif

	int status = templateAnal(card, keyname, keyval, cmt, dtyp) ;
	if ( status) 
	{
		/* we end up here for COMMENT and HEADER and END cards
		* because get_keytype() will set status=204 for these.
		*/
		//fprintf(stderr,__FILE__" %d %s\n",__LINE__,card) ;
		if ( verbose)
			fits_report_error(stderr,status) ;
		return false ;
	}

	int keyvalI ;
	double keyvalD ;
	float keyvalCre, keyvalCim ;

	if (verbose)
		cout << "...appending " << keyname << endl ;
	/* todo: ensure that the restricted classes (BINPIX etc) are not replaced
	*/
	switch( dtyp[0])
	{
	case 'C':
		/* Character keyword. keyval contains already the two delimiting tic marks.
		* Delete them because CCfits will rejoin them.
		*/
		//hdu.addKey(string(keyname),string(keyval),string(cmt)) ;
		hdu.addKey(string(keyname),string(keyval,1,strlen(keyval)-2),string(cmt)) ;
		break ;
	case 'L':
		/* logcal keyword
		*/
		if ( strstr((const char *)keyval,"T") != NULL)
			hdu.addKey(string(keyname),true,string(cmt)) ;
		else
			hdu.addKey(string(keyname),false,string(cmt)) ;
		break ;
	case 'I':
		/* integer keyword
		*/
		sscanf(keyval,"%d",&keyvalI) ;
		hdu.addKey(string(keyname),keyvalI,string(cmt)) ;
		break ;
	case 'F':
		/* floating/double keyword
		*/
		sscanf(keyval,"%lf",&keyvalD) ;
		hdu.addKey(string(keyname),keyvalD,string(cmt)) ;
		break ;
	case 'X':
		/* complex keyword
		*/
		sscanf(keyval,"(%f,%f)",&keyvalCre,&keyvalCim) ;
		complex<float> keyvalC(keyvalCre,keyvalCim) ;
		hdu.addKey(string(keyname),keyvalC,string(cmt)) ;
	}
	return true;
} /* addHduCard */
