/** @file
* csv2Fits translates an ASCII file in the CSV format to a FITS file with a binary table.
*/

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>

#include <cmath>
#include <cstdlib>
#include <cstring>

#include <CCfits/CCfits>

#include "csv2Fits.h"

/** Activate some debugging. 
* Not useful for production.
*/
//#define DEBUG

/** If defined, use two scans of the input CSV file to allocate
* all rows of the binary FITS table in advance 
*/
#define CSV2FITS_WITHLINECNT

using namespace std;
using namespace CCfits;

/** Constructor.
* @param[in] infile The name of the ASCII file in CSV format.
* @param[in] f The field separator. 
*   This is a single character (the comma by default) which separates adjacent fields in
*   the input file.
* @author Richard J. Mathar
* @since 2013-04-30
*/
csv2Fits::csv2Fits(const std::string infile, const char f) : csvFil(infile), fsep(f)
{
}

/** Emit a short usage reminder.
* @param[in] argv0 The name of the executable.
* @author R. J. Mathar
* @since 2013-04-30
*/
void usage(char *argv0)
{
        cout << "Usage:" << endl ;
        cout << argv0 << " [-h] [-F fieldsep] typestring infile.csv" << endl ;
}

/** Determine the number of characters in the longest string in each column.
* @param[out] lineCt On return this contains the number of lines in the input file.
*  The header line with the prospective names of the columns is not counted.
* @return A vector of maximum character widths for columns 0, 1, 2, ... etc
*  This summarizes the widths of the associated column along all lines of the input file.
* @author R. J. Mathar
* @since 2013-04-30
*/
vector<int> csv2Fits::colWidths(int &lineCt) const
{
	vector<int> w ;
	lineCt = 0 ;
	/* open file
	*/
	ifstream ifi(csvFil.c_str()) ;
	if ( ! ifi)
	{	
		cerr << "Could not open " << csvFil << endl ;
	}
	else
	{
		/* loop over all lines in the file
		*/
		int lineno =0 ;
		while( !ifi.eof()) 
		{
			string lin ;
			getline(ifi,lin) ;
			vector<string> fields = strtok(lin) ;
#ifdef DEBUG
#if 0
			/* debugging input fields
			*/
			for(int  i=0 ; i < fields.size() ; i++)
				cout << fields[i] <<":" ;
			cout << endl ;
#endif
#endif
			if ( lineno == 0 )
			{
				/* at the first (header) line: all field widths initally zero
				*/
				w = vector<int>(fields.size()) ;
			}
			else
			{
				for(int f=0 ; f < fields.size() ; f++)
					if ( fields[f].size() > w[f])
						w[f]  = fields[f].size() ;
				lineCt++ ;
			}
			lineno++ ;
		}
	}
#ifdef DEBUG
#if 0
	/* debugging input fields
	*/
	for(int  i=0 ; i < w.size() ; i++)
		cout << w[i] <<"|" ;
	cout << endl ;
#endif
#endif
	return w;
} /* colWidths */

/** Scan all lines of the input file and generate the FITS file.
* @param[in] typs A type string with letters a, d, f, b, i, k, and j
*   indicating the FITS type of each column (each field in the CSV)
* @author R. J. Mathar
* @since 2013-04-30
*/
void csv2Fits::cnvrt(const std::string typs) const
{
	/* open file
	*/
	ifstream ifi(csvFil.c_str()) ;
	if ( ! ifi)
	{	
		cerr << "Could not open " << csvFil << endl ;
	}
	else
	{
		int lineCnt ;
		vector<int> fwid = colWidths(lineCnt) ;

		/* output file name
		*/
		string::size_type suf= csvFil.find(".csv") ;
		string ofi( csvFil.substr(0,suf) + ".fits" );

		FITS *ofit ;
		FITS::setVerboseMode(true) ;
		try
		{
			ofit = new FITS(ofi,Write) ;
		}
		catch (CCfits::FITS::CantOpen)
		{
			cerr << "Could not create " << ofi << endl ;
			return ;
		}

		std::vector<string> coName ;
		std::vector<string> coForm ;
		std::vector<string> coUnit ;

		Table *newTab ;

		/* loop over all lines in the file
		*/
		int lineno =0 ;
		int fieldNo;
		double tmpD ;
		float tmpF ;
		unsigned char tmpB ;
		short tmpI ;
		long long tmpK ;
		int tmpJ ;

		while( !ifi.eof() ) 
		{
			string lin ;
			getline(ifi,lin) ;
			vector<string> fields = strtok(lin) ;
			if ( lineno == 0 )
			{
				/* header line provides the names of the columns */
				coName = fields ;
				/* the typs provides the data types of the columns
				*/
				if ( fields.size() != typs.size())
					cerr << "Warning " << typs.size() << " in " << typs << " but " << fields.size() << " fields in " << csvFil << endl ;

				fieldNo = min( fields.size(), typs.size()) ;
				coName.resize(fieldNo) ;
				for(int f=0 ; f < fieldNo ; f++)
				{
					char Aform[12] ;
					
					switch( typs[f] )
					{
					case 'a':
						snprintf(Aform,11,"%dA",fwid[f]) ;
						coForm.push_back(Aform) ;
						coUnit.push_back("") ;
						break;
					case 'd':
						// coForm.push_back(string("D")+fwid[f]) ;
						coForm.push_back("1D") ;
						coUnit.push_back("") ;
						break;
					case 'f':
						coForm.push_back("1E") ;
						coUnit.push_back("") ;
						break;
					case 'b':
						coForm.push_back("1B") ;
						coUnit.push_back("") ;
						break;
					case 'i':
						coForm.push_back("1I") ;
						coUnit.push_back("") ;
						break;
					case 'k':
						coForm.push_back("1K") ;
						coUnit.push_back("") ;
						break;
					case 'j':
						coForm.push_back("1J") ;
						coUnit.push_back("") ;
						break;
					default:
						cerr << "Invalid type specifier " << typs[f] << endl ;
					}
				}
#ifdef DEBUG
				for(int f=0 ; f < fieldNo ; f++)
					cout << coName[f] << " " << coForm[f] << "|" ;
				cout << endl ;
#endif

				// Table * newTab = ofit->addTable("FILES",10,coName,coForm,coUnit,BinaryTbl) ;
#ifdef CSV2FITS_WITHLINECNT
				newTab = ofit->addTable("FILES",lineCnt,coName,coForm,coUnit) ;
#else
				newTab = ofit->addTable("FILES",0,coName,coForm,coUnit) ;
#endif
#ifdef DEBUG
				cout << lineno << " " << lin << endl ;
#endif
			}
			else
			{
				/* any line which is not a header line
				*/
#if 0
				cout << lineno << " " << lin << endl ;
#endif

#ifndef CSV2FITS_WITHLINECNT
				newTab->insertRows(lineno-1,1) ;
#endif
				fieldNo = min( fields.size(), typs.size()) ;
				for(int f=0 ; f < fieldNo ; f++)
				{
					vector<string> tmpstr ;
					Column & co = newTab ->column(f+1) ;
#ifdef DEBUG
					cerr << typs[f] << endl ;
#endif
					if ( fields[f].length() )
						switch( typs[f] )
						{
						case 'a':
							// co.write(fields[f].c_str(),1L,(long)lineno) ;
							// co.write(fields[f].c_str(),fields[f].size(),1L,(long)lineno) ;
							tmpstr.push_back(fields[f]) ;
							co.write(tmpstr,(long)lineno) ;
							break;
						case 'd':
							if ( fields[f].length() )
							{
								tmpD = atof(fields[f].c_str()) ;
								co.write(&tmpD,1L,(long)lineno) ;
							}
							break;
						case 'f':
							if ( fields[f].length() )
							{
								tmpF = atof(fields[f].c_str()) ;
								co.write(&tmpF,1L,(long)lineno) ;
							}
							break;
						case 'b':
							if ( fields[f].length() )
							{
								tmpB = atoi(fields[f].c_str()) ;
								co.write(&tmpB,1L,(long)lineno) ;
							}
							break;
						case 'i':
							tmpI = atoi(fields[f].c_str()) ;
							co.write(&tmpI,1L,(long)lineno) ;
							break;
						case 'k':
							tmpK = atoi(fields[f].c_str()) ;
							co.write(&tmpK,1L,(long)lineno) ;
							break;
						case 'j':
							tmpJ = atoi(fields[f].c_str()) ;
							co.write(&tmpJ,1L,(long)lineno) ;
							break;
							break;
						default:
							cerr << "Invalid type specifier " << typs[f] << endl ;
						}
				}
			}
			lineno++ ;
		}
		(ofit->pHDU()).writeDate() ;
		delete ofit ;
	}
} /* cnvrt */

/** Decompose the CSV line by splitting it along commas
* @param[in] lin The input line with any number of field separators.
* @return The first field in the 0-th element, the 2nd field in the 1-st element etc.
*/
vector<string> csv2Fits::strtok(const std::string lin) const
{
	vector<string> res ;
	string::size_type strt=0 ;

	string::size_type stop=lin.find_first_of(fsep) ;
	while( stop != string::npos)
	{
		res.push_back( lin.substr(strt,stop-strt) ) ;
		strt = stop+1 ;
		stop=lin.find_first_of(fsep,strt) ;
	}
	if ( stop > strt)
		res.push_back( lin.substr(strt,stop-strt) ) ;
	return res ;
}

/** Scan an ASCII file (CSV format) and convert this into a binary FITS table.
* Given a readable CSV file in the file system, the program generates
* a FITS file which contains a single extension in the binary FITS table format.
*
* The name of the FITS file is chosen by replacing the .csv extension of the
* CSV source file by the extension .fits.
*
* Each line (but the first) of the CSV file becomes a row in the FITS table.
* The comma-separated first line in the CSV file contains the names of the
* columns to be used in the FITS columns.
*
* The syntax in overview:
*
* <code>csv2Fits [-h] [-F fieldsep] typstr infile.csv</code>
*
* The option -h just emits a reminder for the syntax of the call.
* The same reminder is given if the two mandatory command line arguments are missing.
*
* The option -F followed by a single character changes the field separator
* definition in infile.csv. The default is the comma.
*
* The first command line argument is a concatenation of one or more letters
* of the set {a, d, f, b, i, k, j}. There are exactly as many letters as columns
* in the CSV file, and from left to right, each letter denotes the FITS type
* of the associated column of the CSV file. The Makefile shows the example
*
* <code>csv2Fits aaaaadafafbbaa csv2FitsXmpl.csv</code>
*
* with 5 ASCII strings in the first 5 columns, one double value in the 6th column etc.
* - The letter 'a' triggers the xA FITS format for strings.
*   The width is determined by the maximum length of strings in that particular column
*   of the CSV file. In consequence the 'x' in this format will differ for most
*   of the columns that are in the ASCII/string format.
* - The letter 'd' triggers the 1D format for numbers in double-precision.
* - The letter 'f' triggers the 1F format for numbers in single-precision
* - The letter 'b' triggers the 1B format for 8-bit integers.
* - The letter 'i' triggers the 1I format for 16-bit integers.
* - The letter 'j' triggers the 1J format for 32-bit integers.
* - The letter 'k' triggers the 1K format for 64-bit integers.
*
* @example
*   To convert the SAO star catalogue of
*   http://heasarc.gsfc.nasa.gov/FTP/heasarc/dbase/dump/heasarc_sao.tdat.gz
*   into  FITS file, (i) create an ASCII heasarc_sao.csv file by deleting the
*   lines up to and including the <DATA> line, deleting the last <END> line,
*   (ii) insert a first line 
NAME|RA|PM_RA|PM_RA_ERR|RA_EPOCH|DEC|PM_DEC|PM_DEC_ERR|DEC_EPOCH|POS_ERR|LII|BII|PG_MAG|V_MAG|SPECTYP|REF_VMAG|REF_STARNUM|REF_PG_MAG|REF_PM|REF_SPECTYP|REMARKS|REF_SRC_CAT|NUM_SRC_CAT|DM|HD|HD_COMPO|GC|PM_RA_FK5|PM_DEC_FK5|CLASS
*   at the place of the former <DATA> line,
*   and (iii) call csv2Fits -F '|' addbdddbddddddabbbbbbbjaaaaddi heasarc_sao.csv
*   
* @param[in] argc The number of command line arguments.
* @param[in] argv The vector of options, arguments of the options and free-standing arguments
*
* @author Richard J. Mathar
* @since 2013-04-30
*/
int main(int argc, char *argv[])
{
	/* command line option character. Loop over all of the comand line arguments.
	*/
	char oc ;

	char fieldsep = ',' ;

	while ( (oc=getopt(argc,argv,"hF:")) != -1 )
	{
		switch(oc)
		{
		case 'h' :
			usage(argv[0]) ;
			return 0 ;
			break;
		case 'F' :
			fieldsep = optarg[0] ;
			break;
		case '?' :
                        cerr << "Invalid command line option " << optopt << endl ;
                        break ;
		}
	}
#ifdef DEBUG
	cerr << "Field separator " << fieldsep << endl ;
#endif

	if ( optind+1 < argc)
	{
		std::string typs=argv[optind] ;
		std::string infile=argv[optind+1] ;
	
		csv2Fits csv(infile,fieldsep) ;
	
		csv.cnvrt(typs) ;
		return 0 ;
	}
	else
	{
		usage(argv[0]) ;
		return 1 ;
	}

} /* main */
