/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include <unistd.h>

#include "FEditHead.h"

/** Print a usage syntax message detailing the command line
* @param[in] argv0 The name of the main program.
* @since 2012-10-18
*/
void usage(char *argv0)
{
	cout << "Usage:" << endl ;
	cout << argv0 << " [-v] fitsfilename templatehdrfilename [templatehdrfilename ...]" << endl ;
} /* usage */

/** Fedithead is a standalone program which edits FITS header data
* following directions from a configuration file.
*
* An optional argument <code>-v</code> triggers a detailed message
* of the program for each keyword changed.
*
* The first command line argument is the path/file name of an existing
* FITS file which is to be modified.
*
* The second argument and optionally further arguments are ASCII
* files structured very similar to the template files used with cfitsio
* and fmodhead.
*
* Each of these may contain empty lines and comment lines (starting with #)
* that have no effect.
*
* It may contain lines starting with the dash (-) that
* demand removal of the keyword from the FITS header. (If that keyword does
* not exist this does not have  any effect.) The keyword may have regex
* expressions to deal with a group of keywords at once.
*
* It man contain lines that embed
* two keyword names between colons (:) or between exclamation marks (!),
* so there are three of these delimiters in that type
* of line. (This is a syntactical extension to template files of fmodhead, fedithd
* and the cfitsio templates). Fits header cards with names matching the
* regular expression delimited by the first two colons have their names substituted
* by the substitutional expression between the 2nd and third colon. (Values and
* comments remain as they are).
* 
* It may contain lines that start with at least 8 blanks. The rest of
* these lines is turned into COMMENT lines that are appended to the FITS
* header.
*
* Finally, all other lines are interpreted as
* keyword-value-comment triples in FITS header style (with = and / as delimiters),
* that trigger adding that card to the header. (Existing keywords with the same
* name are removed).
*
* Example file (see also the <code>fmodhead</code> examples in the source
* distribution:

* <code> # delete CHOP_A and CHOP_B</code>

* <code> -CHOP_[AB]</code>

* <code> # replace RHUM by a hierarchical version</code>

* <code> :RHUM:HIERARCH LN AMBI RHUM:</code>

* <code> # rename enumerated wheels to filters</code>

* <code> :WHEEL(\1):HIERARCH LN ICS FILT\1:</code>

* <code> # add a OBSERVAT keyword</code>

* <code> OBSERVAT = "LBT" / on the mountain</code>

* <code> # add a comment</code>

* <code>          Nice observation conditions. Dry with occasional snowflakes.</code>

* @param[in] argc The number of command line switches and arguments.
* @param[in] argv The vector of all command line arguments.
* @return 0 if the file manipulation was successful.
* @since 2012-10-18
* @since 2013-10-30 Support more than one template header file.
*/
int main(int argc, char **argv)
{
	bool verbose = false ;
	/* option character
	*/
	char oc ;
        while (  (oc=getopt(argc,argv,"v")) != -1 )
        {
                switch(oc)
                {
                case 'v' :
                        verbose = true ;
                        break ;
                case '?' :
                        cerr << "Invalid command line option " << optopt << endl ;
                        break ;
                }
        }

        /* ensure that this is called with at least two file names
        */
        if ( optind +2 > argc )
        {
                usage(argv[0]) ;
                return 1 ;
        }

	vector<string> fitstpl ;
	for(int f= optind+1 ; f < argc ; f++)
		fitstpl.push_back(string(argv[f])) ;	 

	FEditHead fedt(argv[optind],fitstpl) ;
	fedt.exec(verbose) ;

	return 0 ;
} /* main */
